/*
Firmware for an ATTiny85 to act as the controller for a LiIon psu that
can either charge the batteries from the 5V applied to it or generate
5V from the batteries for e.g. a Raspberry Pi
(C) 2012 Jeroen Domburg (jeroen AT spritesmods.com)

This program is free software: you can redistribute it and/or modify
t under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#define F_CPU 8000000

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdio.h>


#define ENA5V PB0 //also: Vref to the battery
#define PWMBUCK PB1 //OC1A
#define ANASHUNT PB3
#define PWMBOOST PB4 //OC1B
#define BTNLED PB2


//Play dead until the wdt wakes us.
void powerdown() {
	//Disable PWM, set outputs to idle levels
	TCCR1=0x00; GTCCR=0x00;
	DDRB=(1<<PWMBUCK);
	PORTB=(1<<BTNLED);

	//Sleepy times now.
	cli();
	wdt_enable(WDTO_1S);
	set_sleep_mode(SLEEP_MODE_PWR_DOWN);
	sleep_mode();
	//won't come back here; uC resets in 1sec.
}


#define LED_OFF				0
#define LED_CHG_CURRENT		1
#define LED_CHG_VOLTAGE		2
#define LED_CHG_DONE		3
#define LED_ON				4

#define BTN_UNPRESSED		0
#define BTN_PRESSED_SHORT	1
#define BTN_PRESSED_LONG	2

#define STATE_PWROFF		0
#define STATE_CHARGE		1
#define STATE_GEN5V			2

volatile char ledstate=0;
volatile char btnstate=0;
char state=0;

//Handle the LEDs and long button presses
//Called about 30 times a second.
ISR(TIM0_OVF_vect) {
	static int t=0;
	static char ledpos=0;
	static int buttonPressed=100; //assume we start with the button pressed
	t++;
	if (t>=8) { //4x a second
		//Disable LED for a short while so we can check the button state in the next state
		DDRB&=~(1<<BTNLED);
		PORTB|=(1<<BTNLED);
		_delay_us(100);
		if ((PINB&(1<<BTNLED))==0) buttonPressed++; else buttonPressed=0;
		if (buttonPressed==(1*4)) btnstate=BTN_PRESSED_SHORT;
		if ((buttonPressed==(3*4)) && btnstate==BTN_PRESSED_SHORT) btnstate=BTN_PRESSED_LONG; //force powerdown if button pressed for >3sec
		t=0;
		ledpos=(ledpos+1)&3;
		if (ledstate>ledpos) {
			//Re-enable led, if needed
			PORTB&=~(1<<BTNLED);
			DDRB|=(1<<BTNLED);
		}
	}
}


#define TARGETADC5V 225 //(1.1V/5V)*1024 = 225.
#define PSCALE5V 2
#define ITIME5V (256L)
#define ISCALE5V 8

/*
5V generation: The AVR will use the mosfet connected to the PWMBOOST output to convert the 3.6V LiIon voltage
to a 5V voltage the Pi can run on. The AVR is connected to the same 5V line, so we can measure Vcc to adjust
the PWM value to keep the 5V line steady. To measure Vcc we use the a/d-converter: we measure the internal 1.1V
against Vcc as a reference to get the adjustment value. Adjustment goes according to a bit of
PI-like controlling logic.
*/
void gen5v() {
	int offset=-255;
	long longtermval=0;
	int slowstart=0;
	int t=0;
	//Disable buck pwm mosfet so no 5V can leak in over USB
	PORTB|=(1<<PWMBUCK);
	DDRB|=(1<<PWMBUCK);

	TCCR1=0x01; //no pwm on chana, no div
	GTCCR=0x60; //pwm on chanb
	OCR1A=0; OCR1B=0;
	ledstate=LED_OFF; //so the R-Pi can see the button presses too
	PORTB|=(1<<ENA5V); //disable for now
	DDRB|=(1<<ENA5V);


	while(1) {
		int val;
		int comp;
		ADMUX=0x0C; //Vcc ref, sample Vbg
		ADCSRA=0xC3; //adc enable & start, div=/8
		while (ADCSRA&(1<<6)) ;
		val=ADC;
		longtermval+=val;
		//Calculate compensation vale. It's basically the scaled square difference between
		//the current and target ADC value.
		comp=(val-TARGETADC5V)*PSCALE5V;
		comp=(comp>0)?comp*comp:-comp*comp;
		val=comp+offset;
		//clamp
		if (val<=0) val=0;
		if (val>0xA0) val=0xA0;
		if (val>slowstart) val=slowstart;
		OCR1B=val;
		//long-term stuff (the i-part of the pi-regulator)
		t=t+1;
		if (t>=ITIME5V) {
			t=0;
			if (slowstart<255) {
				slowstart++;
			} else {
				PORTB&=~(1<<ENA5V); //enable Pi
			}
			longtermval=longtermval/ITIME5V;
			offset=offset+((longtermval-TARGETADC5V)/ISCALE5V);
		}
		wdt_reset();
		if (btnstate==BTN_PRESSED_LONG) {
			//Go power off.
			state=STATE_PWROFF;
			btnstate=BTN_UNPRESSED;
			return;
		}
	}
}

//LiIon voltage should be limited to 4.2V
#define TARGETADCCHGVCC 261 //(1.1V/4.2V)*1024 = 267.
//Rshunt=0.25 ohm, limit current to 500mA -> limit Vshunt to 125mV
#define TARGETADCCHGSHUNT 116 //(.125/1.1V)*1024 = 116.

/*
The device is connected to the 5V power supply. Use the MosFET connected to PWMBUCK to generate the
voltage to charge the battery. The voltage over the battery can be measured because it's the same
as the Vref of the controller, which can be measured by sticking the ADC on the internal 1.1V reference. 
The current can be measured through the shunt connected to ANASHUNT.
*/
void charge() {
	int x;
	PORTB&=~(1<<PWMBOOST); //will fuck stuff up if high
	DDRB|=(1<<PWMBUCK)|(1<<PWMBOOST); //needed
	TCCR1=0x71; //inverted pwm on chana, no div
	GTCCR=0x00; //no pwm on chanb
	OCR1A=3; OCR1B=0; OCR1C=255;

	//Disable 5V enable line by floating it. Vbat will pull it up.
	DDRB&=~(1<<ENA5V);
	PORTB&=~(1<<ENA5V);

	int pwmval=0;
	while(1) {
		int voltval, shuntval;

		//Don't ask about the loops... I basically ahd trouble getting a correct ADC-value
		//after switching around the muxes, so I just run the DA 10 times. That works, but
		//isn't pretty by a long shot.
		for (x=0; x<10;x++) {
			ADMUX=0x4C; //Vref ref, sample Vbg
			ADCSRA=0xC3; //adc enable & start, div=/8
			while (ADCSRA&(1<<6)) ;
			_delay_ms(2);
			ADCSRA=0xC3; //adc enable & start, div=/8
			while (ADCSRA&(1<<6)) ;
			voltval=ADC;
		}

		for (x=0; x<10;x++) {
			ADMUX=0x83; //1.1V ref, sample Vshunt
			ADCSRA=0xC3; //adc enable & start, div=/8
			while (ADCSRA&(1<<6)) ;
			_delay_ms(2);
			ADCSRA=0xC3; //adc enable & start, div=/8
			while (ADCSRA&(1<<6)) ;
			shuntval=ADC;
		}

		//Set led to something indicating battery state.
		if (voltval<TARGETADCCHGVCC) ledstate=LED_CHG_VOLTAGE;
		if (shuntval>TARGETADCCHGSHUNT) ledstate=LED_CHG_CURRENT;
		if (shuntval<20) ledstate=LED_CHG_DONE;

		//Modify buck regulator value
		if (voltval<TARGETADCCHGVCC || shuntval>TARGETADCCHGSHUNT) pwmval--; else pwmval++;
		if (pwmval<0) pwmval=0;
		if (pwmval>0xff) pwmval=0xff;
		OCR1A=pwmval; OCR1B=pwmval;

		wdt_reset();
		if (pwmval>0xf0 && shuntval<20) {
			//We PWM so much we connect the full 5V to the battery, but get no current. Probably
			//the charger (or maybe the battery) has disconnected.
			state=STATE_PWROFF;
			return;
		}
		if (btnstate!=BTN_UNPRESSED) {
			//Break off charge and start up 5V gen
			btnstate=BTN_UNPRESSED;
			state=STATE_GEN5V;
			return;
		}
	}
}

EMPTY_INTERRUPT(INT0_vect);

//Main routine
int main(void) {
	char btnpressed=0;
	char usbconnected=0;
	wdt_enable(WDTO_1S);

	ACSR=0;
	DDRB=(1<<PWMBOOST)|(1<<BTNLED);
	PORTB=(1<<BTNLED); //Pull button high to overcome the grounding via the RPi
	if (!(ACSR&(1<<5))) usbconnected=1;
	if ((PINB&(1<<BTNLED))==0) btnpressed=1;
	DDRB&=~(1<<BTNLED);
	if (!btnpressed && !usbconnected) powerdown();

	//Ok, we should do something.

	//Enable timer0
	TCCR0A=0x00;
	TCCR0B=5; //src=clk/1024 -> overflows 300 times a second at 8MHz
	TIMSK=2; //enable overflow int

	//Enable PLL
	PLLCSR=0x2;
	_delay_us(100);
	while((PLLCSR&1)==0) ;
	PLLCSR=0x6;

	//Enable ints
	sei();

	if (usbconnected) state=STATE_CHARGE;
	if (btnpressed) state=STATE_GEN5V;

	while(1) {
		//Choose what to do
		if (state==STATE_CHARGE) {
			charge();
		} else if (state==STATE_GEN5V) {
			gen5v();
		} else if (state==STATE_PWROFF) {
			powerdown();
		}
	}

	return 0;
}
